from django.conf import settings

INCIDENT_REPORT_DIALOG = "incident-report-dialog"
INCIDENT_EDIT_DIALOG = "incident-edit-dialog"

SLACK_API_MOCK = getattr(settings, "SLACK_API_MOCK", None)

if SLACK_API_MOCK:
    from urllib.parse import urlparse
    from slackclient.slackrequest import requests

    old_post = requests.post

    import logging

    def fake_post(url, *args, **kwargs):
        parsed = urlparse(url)
        newurl = parsed._replace(netloc=SLACK_API_MOCK, scheme="http")
        logging.info(f"Fake Slack client request: HTTP POST {SLACK_API_MOCK} {newurl}")
        return old_post(newurl.geturl(), *args, **kwargs)

    requests.post = fake_post



import hashlib
import hmac
import logging
from functools import wraps
from time import time

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.handlers.wsgi import WSGIRequest

logger = logging.getLogger(__name__)

slack_signing_secret = settings.SLACK_SIGNING_SECRET


def slack_authenticate(f):
    """
    slack_authenticate is a wrapper to check the request signature.

    usage:

    @slack_authenticate
    def slack_actions():
        pass
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        request = args[0]
        if len(args) < 1 or not type(args[0]) == WSGIRequest:
            logger.error(
                "slack_authenticate annotation used with incorrect args or no args"
            )
            raise ValueError
        if authenticate(request):
            return f(*args, **kwargs)
        else:
            logger.error(f"Denied access to {request.path}")
            raise PermissionDenied

    return wrapper


def authenticate(request):
    """
    authenticate authenticates the request using the signature process provided
    by slack

    see: https://api.slack.com/docs/verifying-requests-from-slack
    """

    if slack_signing_secret == "":
        logger.critical("Signing secret isn't defined")
        return False

    if "HTTP_X_SLACK_REQUEST_TIMESTAMP" not in request.META:
        logger.error("HTTP_X_SLACK_REQUEST_TIMESTAMP not provided")
        return False

    req_timestamp = request.META["HTTP_X_SLACK_REQUEST_TIMESTAMP"]
    if abs(time() - int(req_timestamp)) > 60 * 5:
        logger.error("Invalid request timestamp")
        return False

    if (
        "HTTP_X_SLACK_SIGNATURE" not in request.META
        or request.META["HTTP_X_SLACK_SIGNATURE"] == ""
    ):
        logger.error("HTTP_X_SLACK_SIGNATURE not provided")
        return False

    req_signature = request.META["HTTP_X_SLACK_SIGNATURE"]
    if not verify_signature(
        req_timestamp, req_signature, slack_signing_secret, request.body
    ):
        logger.error("Invalid request signature")
        return False

    return True


def verify_signature(timestamp, signature, secret, data):
    # Verify the request signature of the request sent from Slack
    # Generate a new hash using the app's signing secret and request data

    # Compare the generated hash and incoming request signature
    request_hash = generate_signature(timestamp, secret, data)
    return hmac.compare_digest(request_hash, signature)


def generate_signature(timestamp, secret, data):
    req = str.encode("v0:" + str(timestamp) + ":") + data
    return "v0=" + hmac.new(str.encode(secret), req, hashlib.sha256).hexdigest()



import datetime

import pytest
from faker import Faker

from response.core.models import TimelineEvent
from response.slack.event_handlers import handle_pin_added, handle_pin_removed
from response.slack.models import PinnedMessage
from tests.factories import ExternalUserFactory, IncidentFactory

faker = Faker()


@pytest.mark.django_db
def test_add_and_remove_pin(mock_slack):
    incident = IncidentFactory.create()
    user = ExternalUserFactory.create()

    mock_slack.get_user_profile.return_value = {"name": user.display_name}

    text = faker.paragraph(nb_sentences=2, variable_nb_sentences=True)
    handle_pin_added(
        incident,
        {"item": {"message": {"user": user.external_id, "ts": 123, "text": text}}},
    )
    mock_slack.get_user_profile.assert_called_with(user.external_id)

    message = PinnedMessage.objects.get(incident=incident, message_ts=123)
    assert message.text == text

    handle_pin_removed(
        incident,
        {"item": {"message": {"user": user.external_id, "ts": 123, "text": text}}},
    )

    mock_slack.get_user_profile.assert_called_with(user.external_id)

    with pytest.raises(PinnedMessage.DoesNotExist):
        PinnedMessage.objects.get(incident=incident, message_ts=123)
        TimelineEvent.objects.get(
            incident=incident, timestamp=datetime.fromtimestamp(123)
        )



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import user_passes_test


def response_login_required(
    function=None, redirect_field_name=REDIRECT_FIELD_NAME, login_url=None
):
    """
    Re-implementation of Django's login_required decorator to extend with
    support for allowing anonymous viewing if `RESPONSE_LOGIN_REQUIRED` is True.
    """
    actual_decorator = user_passes_test(
        lambda u: u.is_authenticated or not settings.RESPONSE_LOGIN_REQUIRED,
        login_url=login_url,
        redirect_field_name=redirect_field_name,
    )
    if function:
        return actual_decorator(function)
    return actual_decorator



from django.contrib import admin

from response.slack.models import (
    CommsChannel,
    HeadlinePost,
    Notification,
    PinnedMessage,
    UserStats,
)

# Register your models here.
admin.site.register(HeadlinePost)
admin.site.register(CommsChannel)
admin.site.register(Notification)
admin.site.register(UserStats)
admin.site.register(PinnedMessage)



from django.db import models

from response.core.models import Incident


class Notification(models.Model):
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    key = models.CharField(max_length=30)
    time = models.DateTimeField()
    repeat_count = models.IntegerField(default=0)
    completed = models.BooleanField(default=False)

    class Meta:
        unique_together = ("incident", "key")

    def __str__(self):
        nice_date = self.time.strftime("%Y-%m-%d %H:%M:%S")
        return f"{nice_date} - {self.incident} - {self.key}"



from django.apps import AppConfig
from django.conf import settings as site_settings


class ResponseConfig(AppConfig):
    name = "response"

    def ready(self):
        from .slack import (  # noqa: F401
            settings,
            signals,
            action_handlers,
            event_handlers,
            incident_commands,
            incident_notifications,
            dialog_handlers,
        )

        site_settings.RESPONSE_LOGIN_REQUIRED = getattr(
            site_settings, "RESPONSE_LOGIN_REQUIRED", True
        )



from datetime import datetime

from django.conf import settings

from response.core.models import Action, ExternalUser, Incident
from response.slack.client import SlackError, reference_to_id
from response.slack.decorators.incident_command import (
    __default_incident_command,
    get_help,
)
from response.slack.models import CommsChannel


@__default_incident_command(["help"], helptext="Display a list of commands and usage")
def send_help_text(incident: Incident, user_id: str, message: str):
    return True, get_help()


@__default_incident_command(
    ["summary"], helptext="Provide a summary of what's going on"
)
def update_summary(incident: Incident, user_id: str, message: str):
    incident.summary = message
    incident.save()
    return True, None


@__default_incident_command(["impact"], helptext="Explain the impact of this")
def update_impact(incident: Incident, user_id: str, message: str):
    incident.impact = message
    incident.save()
    return True, None


@__default_incident_command(["lead"], helptext="Assign someone as the incident lead")
def set_incident_lead(incident: Incident, user_id: str, message: str):
    assignee = reference_to_id(message) or user_id
    name = settings.SLACK_CLIENT.get_user_profile(assignee)["name"]
    user, _ = ExternalUser.objects.get_or_create_slack(
        external_id=assignee, display_name=name
    )
    incident.lead = user
    incident.save()
    return True, None


@__default_incident_command(["severity", "sev"], helptext="Set the incident severity")
def set_severity(incident: Incident, user_id: str, message: str):
    for sev_id, sev_name in Incident.SEVERITIES:
        # look for sev name (e.g. critical) or sev id (1)
        if (sev_name in message) or (sev_id in message):
            incident.severity = sev_id
            incident.save()
            return True, None

    return False, None


@__default_incident_command(["rename"], helptext="Rename the incident channel")
def rename_incident(incident: Incident, user_id: str, message: str):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        comms_channel.rename(message)
    except SlackError:
        return (
            True,
            "👋 Sorry, the channel couldn't be renamed. Make sure that name isn't taken already.",
        )
    return True, None


@__default_incident_command(
    ["duration"], helptext="How long has this incident been running?"
)
def set_duration(incident: Incident, user_id: str, message: str):
    duration = incident.duration()

    comms_channel = CommsChannel.objects.get(incident=incident)
    comms_channel.post_in_channel(f"The incident has been running for {duration}")

    return True, None


@__default_incident_command(["close"], helptext="Close this incident.")
def close_incident(incident: Incident, user_id: str, message: str):
    comms_channel = CommsChannel.objects.get(incident=incident)

    if incident.is_closed():
        comms_channel.post_in_channel(
            f"This incident was already closed at {incident.end_time.strftime('%Y-%m-%d %H:%M:%S')}"
        )
        return True, None

    incident.end_time = datetime.now()
    incident.save()

    comms_channel.post_in_channel(f"This incident has been closed! 📖 -> 📕")

    return True, None


@__default_incident_command(["action"], helptext="Log a follow up action")
def set_action(incident: Incident, user_id: str, message: str):
    name = settings.SLACK_CLIENT.get_user_profile(user_id)["name"]
    action_reporter, _ = ExternalUser.objects.get_or_create_slack(
        external_id=user_id, display_name=name
    )
    Action(incident=incident, details=message, user=action_reporter).save()
    return True, None



import logging
from datetime import datetime

from response.core.models import Incident
from response.slack.models.notification import Notification

logger = logging.getLogger(__name__)


NOTIFICATION_HANDLERS = []


class NotificationHandler(object):
    def __init__(self, key, callback, interval_mins, max_notifications):
        self.key = key
        self.callback = callback
        self.interval_mins = interval_mins
        self.max_notifications = max_notifications

    def __str__(self):
        return self.key


def single_notification(initial_delay_mins=0, func=None):
    """
    Register a handler that'll be called once in each open incident
    """

    def _wrapper(fn):
        NOTIFICATION_HANDLERS.append(
            NotificationHandler(
                key=fn.__name__,
                callback=fn,
                interval_mins=initial_delay_mins,
                max_notifications=0,
            )
        )
        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def recurring_notification(interval_mins, max_notifications=1):
    """
    Register a handler that'll be called periodically for all open incidents.
    """

    def _wrapper(fn):
        NOTIFICATION_HANDLERS.append(
            NotificationHandler(
                key=fn.__name__,
                callback=fn,
                interval_mins=interval_mins,
                max_notifications=max_notifications - 1,
            )
        )
        return fn

    return _wrapper


def handle_notifications():

    # Only notify open incidents with a comms channel
    open_incidents = Incident.objects.filter(
        end_time__isnull=True, commschannel__incident__isnull=False
    )

    for incident in open_incidents:
        for handler in NOTIFICATION_HANDLERS:
            try:
                # get the last sent notification for this incident/handler pair
                notification = Notification.objects.get(
                    incident=incident, key=handler.key
                )

                # if we can find a previous notification for this incident/handler pair and
                # it's been set to not repeat, exit here
                if (
                    notification.repeat_count >= handler.max_notifications
                    or notification.completed
                ):
                    continue

                # it's not exhausted its max_notifications, so wait 'interval_mins' before sending again
                mins_since_last_notify = int(
                    (datetime.now() - notification.time).total_seconds() / 60
                )
                if mins_since_last_notify >= handler.interval_mins:
                    try:
                        handler.callback(incident)
                    except Exception as e:
                        logger.error(
                            f"Error calling notification handler {handler}: {e}"
                        )

                    notification.time = datetime.now()
                    notification.repeat_count = notification.repeat_count + 1
                    notification.save()

            except Notification.DoesNotExist:
                # we've never sent a notification to this incident/handler pair,
                # so wait until 'interval_mins' mins have elapsed from start
                mins_since_started = int(
                    (datetime.now() - incident.start_time).total_seconds() / 60
                )
                if mins_since_started >= handler.interval_mins:
                    try:
                        handler.callback(incident)
                    except Exception as e:
                        logger.error(
                            f"Error calling notification handler {handler}: {e}"
                        )

                    notification = Notification(
                        incident=incident,
                        key=handler.key,
                        time=datetime.now(),
                        repeat_count=0,
                    )
                    notification.save()



from django.db import models

from response.core.models.incident import Incident
from response.core.models.user_external import ExternalUser


class Action(models.Model):
    details = models.TextField(blank=True, default="")
    done = models.BooleanField(default=False)
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    user = models.ForeignKey(
        ExternalUser, on_delete=models.CASCADE, blank=False, null=False
    )

    def icon(self):
        return "🔜️"

    def __str__(self):
        return f"{self.details}"



import factory
from faker import Factory

from response.core.models import Action

faker = Factory.create()


class ActionFactory(factory.DjangoModelFactory):
    class Meta:
        model = Action

    user = factory.SubFactory("tests.factories.ExternalUserFactory")

    details = factory.LazyFunction(
        lambda: faker.paragraph(nb_sentences=1, variable_nb_sentences=True)
    )

    done = factory.LazyFunction(lambda: faker.boolean(chance_of_getting_true=25))



from .action import ActionFactory
from .incident import IncidentFactory
from .timeline import TimelineEventFactory
from .user import ExternalUserFactory, UserFactory

__all__ = (
    "IncidentFactory",
    "TimelineEventFactory",
    "UserFactory",
    "ActionFactory",
    "ExternalUserFactory",
)



import json

import pytest
from django.urls import reverse
from faker import Faker
from rest_framework.test import force_authenticate

from response import serializers
from response.core.views import IncidentTimelineEventViewSet
from response.models import TimelineEvent
from tests.factories import IncidentFactory, TimelineEventFactory

faker = Faker()


def test_create_timeline_event(arf, api_user):
    incident = IncidentFactory.create()

    event_model = TimelineEventFactory.build(incident=incident)
    event_data = serializers.TimelineEventSerializer(event_model).data

    req = arf.post(
        reverse("incident-timeline-event-list", kwargs={"incident_pk": incident.pk}),
        event_data,
        format="json",
    )
    force_authenticate(req, user=api_user)
    response = IncidentTimelineEventViewSet.as_view({"post": "create"})(
        req, incident_pk=incident.pk
    )

    assert response.status_code == 201, "Got non-201 response from API"

    assert TimelineEvent.objects.filter(
        incident=incident, timestamp=event_model.timestamp
    ).exists()


def test_list_timeline_events_by_incident(arf, api_user):
    incident = IncidentFactory.create()

    req = arf.get(
        reverse("incident-timeline-event-list", kwargs={"incident_pk": incident.pk})
    )
    force_authenticate(req, user=api_user)
    response = IncidentTimelineEventViewSet.as_view({"get": "list"})(
        req, incident_pk=incident.pk
    )

    content = json.loads(response.rendered_content)
    print(content)
    assert response.status_code == 200, "Got non-200 response from API"

    assert len(content["results"]) == len(incident.timeline_events())
    for event in content["results"]:
        assert event["timestamp"]
        assert event["text"]
        assert event["event_type"]
        assert event["id"]
        if event["event_type"] != "text":
            assert event["metadata"]
            assert type(event["metadata"]) == dict


@pytest.mark.parametrize(
    "update_key,update_value",
    (
        ("", ""),  # no update
        (
            "timestamp",
            faker.date_time_between(start_date="-3d", end_date="now", tzinfo=None),
        ),
        ("text", faker.paragraph(nb_sentences=5, variable_nb_sentences=True)),
    ),
)
def test_update_timeline_event(arf, api_user, update_key, update_value):
    incident = IncidentFactory.create()

    event_model = incident.timeline_events()[0]
    event_data = serializers.TimelineEventSerializer(event_model).data

    if update_key:
        event_data[update_key] = update_value

    req = arf.put(
        reverse("incident-timeline-event-list", kwargs={"incident_pk": incident.pk}),
        event_data,
        format="json",
    )
    force_authenticate(req, user=api_user)
    response = IncidentTimelineEventViewSet.as_view({"put": "update"})(
        req, incident_pk=incident.pk, pk=event_model.pk
    )

    assert response.status_code == 200, "Got non-200 response from API"
    if update_key:
        new_event = TimelineEvent.objects.get(pk=event_model.pk)
        assert (
            getattr(new_event, update_key) == update_value
        ), "Updated value wasn't persisted to the DB"


def test_delete_timeline_event(arf, api_user):
    incident = IncidentFactory.create()

    event_model = incident.timeline_events()[0]

    req = arf.delete(
        reverse("incident-timeline-event-list", kwargs={"incident_pk": incident.pk})
    )
    force_authenticate(req, user=api_user)
    response = IncidentTimelineEventViewSet.as_view({"delete": "destroy"})(
        req, incident_pk=incident.pk, pk=event_model.pk
    )

    assert response.status_code == 204, "Got non-204 response from API"
    with pytest.raises(TimelineEvent.DoesNotExist):
        TimelineEvent.objects.get(pk=event_model.pk)



import json

import pytest
from django.urls import reverse
from faker import Faker
from rest_framework.test import force_authenticate

from response import serializers
from response.core.views import IncidentActionViewSet
from response.models import Action
from tests.factories import ActionFactory, ExternalUserFactory, IncidentFactory

faker = Faker()


def test_list_actions_by_incident(arf, api_user):
    incident = IncidentFactory.create()

    req = arf.get(reverse("incident-action-list", kwargs={"incident_pk": incident.pk}))
    force_authenticate(req, user=api_user)
    response = IncidentActionViewSet.as_view({"get": "list"})(
        req, incident_pk=incident.pk
    )

    content = json.loads(response.rendered_content)
    print(content)
    assert response.status_code == 200, "Got non-200 response from API"

    assert len(content["results"]) == len(incident.action_items())
    for action in content["results"]:
        assert action["details"]
        assert "done" in action
        assert action["user"]


def test_create_action(arf, api_user):
    incident = IncidentFactory.create()
    user = ExternalUserFactory.create()

    action_model = ActionFactory.build(user=user)
    action = serializers.ActionSerializer(action_model).data

    req = arf.post(
        reverse("incident-action-list", kwargs={"incident_pk": incident.pk}),
        action,
        format="json",
    )
    force_authenticate(req, user=api_user)
    response = IncidentActionViewSet.as_view({"post": "create"})(
        req, incident_pk=incident.pk
    )

    assert response.status_code == 201, "Got non-201 response from API"
    assert Action.objects.filter(details=action_model.details).exists()


def test_update_action_user(arf, api_user):
    incident = IncidentFactory.create()
    user = ExternalUserFactory.create()

    action = incident.action_items()[0]

    action_data = serializers.ActionSerializer(action).data
    action_data["user"] = {
        "app_id": "slack",
        "display_name": user.display_name,
        "external_id": user.external_id,
        "full_name": user.full_name,
    }
    response = update_action(arf, api_user, incident.pk, action_data)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-201 response from API"

    updated_action = Action.objects.get(pk=action.pk)
    assert updated_action.user == user


def test_update_action(arf, api_user):
    incident = IncidentFactory.create()

    action = incident.action_items()[0]

    action_data = serializers.ActionSerializer(action).data

    action_data["details"] = faker.paragraph(nb_sentences=2, variable_nb_sentences=True)
    response = update_action(arf, api_user, incident.pk, action_data)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-201 response from API"

    updated_action = Action.objects.get(pk=action.pk)
    assert updated_action.details == action_data["details"]

    action_data["done"] = not action_data["done"]
    response = update_action(arf, api_user, incident.pk, action_data)
    print(response.rendered_content)
    assert response.status_code == 200, "Got non-201 response from API"

    updated_action = Action.objects.get(pk=action.pk)
    assert updated_action.done == action_data["done"]


def update_action(arf, api_user, incident_id, action_data):
    req = arf.put(
        reverse("incident-action-list", kwargs={"incident_pk": incident_id}),
        action_data,
        format="json",
    )
    force_authenticate(req, user=api_user)

    return IncidentActionViewSet.as_view({"put": "update"})(
        req, incident_pk=incident_id, pk=action_data["id"]
    )


def test_delete_action(arf, api_user):
    incident = IncidentFactory.create()
    user = ExternalUserFactory.create()

    action = ActionFactory.create(user=user, incident=incident)

    req = arf.delete(
        reverse("incident-action-list", kwargs={"incident_pk": incident.pk}),
        format="json",
    )
    force_authenticate(req, user=api_user)
    response = IncidentActionViewSet.as_view({"delete": "destroy"})(
        req, incident_pk=incident.pk, pk=action.pk
    )

    assert response.status_code == 204, "Got non-204 response from API"
    with pytest.raises(Action.DoesNotExist):
        Action.objects.get(pk=action.pk)



import json
import re

from response.core.models.incident import Incident
from response.slack.decorators import (
    handle_incident_command,
    handle_keywords,
    slack_event,
)
from response.slack.models import PinnedMessage, UserStats


def decode_app_mention(payload):
    """
    Convenience function for dealing with the common pattern for app mention slack events.

    Example: @bot create Something's happened
    Extracts:  slack_id -> "@bot"
               command  -> "create"
               extra    -> "Something's happened"

    @param payload an app mention string of the form @bot create Something's happened
    @return: three variables split by bot, action and extra text
    """
    incident_command_re = re.compile(r"(<@[A-Z0-9]+>)\s+(\w+)\s*(.*)")
    m = incident_command_re.match(payload.strip())

    if not m or len(m.groups()) != 3:
        raise ValueError(
            f'Could not parse app mention payload "{payload}" into "@<bot> <command> <text>"'
        )

    slack_id = m.group(1)
    command = m.group(2)
    extra = m.group(3)

    return slack_id, command, extra


@slack_event("app_mention")
def handle_app_mention(incident: Incident, payload: json):
    """
    Events of type `app_mention` are of the form `@incident some text here`
    We handle all of these as incident_commands
    """
    _, command_name, message = decode_app_mention(payload["text"])

    channel_id = payload["channel"]
    user_id = payload["user"]
    thread_ts = payload["ts"]

    handle_incident_command(command_name, message, thread_ts, channel_id, user_id)


@slack_event("message")
def handle_keyword_detection(incident: Incident, payload: json):
    handle_keywords(incident, payload)


@slack_event("message")
def update_user_stats(incident: Incident, payload: json):
    # message with a user field are messages coming from users.
    if "user" in payload:
        user_id = payload["user"]
        UserStats.increment_message_count(incident, user_id)


@slack_event("pin_added")
def handle_pin_added(incident, payload):
    """
    Handler for pin added events

    @param payload a slack pin_added event
    """
    pinned_message = payload["item"]["message"]

    if "user" in pinned_message:
        author_id = pinned_message["user"]
        message_ts = pinned_message["ts"]
        text = pinned_message["text"]
        PinnedMessage.objects.add_pin(incident, message_ts, author_id, text)


@slack_event("pin_removed")
def handle_pin_removed(incident, payload):
    """
    Handler for pin removed events

    @param payload a slack pin_removed event
    """

    pinned_message = payload["item"]["message"]
    message_ts = pinned_message["ts"]

    PinnedMessage.objects.remove_pin(incident, message_ts)



import random

import factory
from django.db.models.signals import post_save
from faker import Factory

from response.core.models import Incident
from response.slack.models import CommsChannel

from .action import ActionFactory

faker = Factory.create()


class CommsChannelFactory(factory.DjangoModelFactory):
    class Meta:
        model = CommsChannel


@factory.django.mute_signals(post_save)
class IncidentFactory(factory.DjangoModelFactory):
    class Meta:
        model = Incident

    impact = factory.LazyFunction(
        lambda: faker.paragraph(nb_sentences=1, variable_nb_sentences=True)
    )
    report = factory.LazyFunction(
        lambda: faker.paragraph(nb_sentences=3, variable_nb_sentences=True)
    )
    report_time = factory.LazyFunction(
        lambda: faker.date_time_between(start_date="-6m", end_date="now", tzinfo=None)
    )

    reporter = factory.SubFactory("tests.factories.ExternalUserFactory")
    lead = factory.SubFactory("tests.factories.ExternalUserFactory")

    start_time = factory.LazyFunction(
        lambda: faker.date_time_between(start_date="-6m", end_date="now", tzinfo=None)
    )

    if random.random() > 0.5:
        end_time = factory.LazyAttribute(
            lambda a: faker.date_time_between(start_date=a.start_time, end_date="now")
        )

    severity = factory.LazyFunction(lambda: str(random.randint(1, 4)))
    summary = factory.LazyFunction(
        lambda: faker.paragraph(nb_sentences=3, variable_nb_sentences=True)
    )

    related_channel = factory.RelatedFactory(CommsChannelFactory, "incident")
    related_action_items = factory.RelatedFactoryList(
        ActionFactory, "incident", size=lambda: random.randint(1, 5)
    )
    related_timeline_events = factory.RelatedFactoryList(
        "tests.factories.TimelineEventFactory",
        "incident",
        size=lambda: random.randint(1, 20),
    )



from rest_framework import serializers

from response.core.models import Action, ExternalUser, Incident, TimelineEvent
from response.slack.models import CommsChannel


class ExternalUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExternalUser
        fields = ("app_id", "external_id", "display_name", "full_name")


class TimelineEventSerializer(serializers.ModelSerializer):
    metadata = serializers.JSONField(allow_null=True)

    class Meta:
        model = TimelineEvent
        fields = ("id", "timestamp", "text", "event_type", "metadata")
        read_only_fields = ("id",)


class ActionSerializer(serializers.ModelSerializer):
    user = ExternalUserSerializer()

    class Meta:
        model = Action
        fields = ("id", "details", "done", "user")
        read_only_fields = ("id",)

    def create(self, validated_data):
        user = ExternalUser.objects.get(
            app_id=validated_data["user"]["app_id"],
            display_name=validated_data["user"]["display_name"],
            external_id=validated_data["user"]["external_id"],
            full_name=validated_data["user"]["full_name"],
        )
        validated_data["user"] = user
        return Action.objects.create(**validated_data)

    def update(self, instance, validated_data):
        if "user" in validated_data:
            instance.user = ExternalUser.objects.get(
                app_id=validated_data["user"]["app_id"],
                display_name=validated_data["user"]["display_name"],
                external_id=validated_data["user"]["external_id"],
                full_name=validated_data["user"]["full_name"],
            )
        instance.details = validated_data.get("details", instance.details)
        instance.done = validated_data.get("done", instance.done)
        instance.save()
        return instance


class CommsChannelSerializer(serializers.ModelSerializer):
    class Meta:
        model = CommsChannel
        fields = ("channel_id", "channel_name")


class IncidentSerializer(serializers.ModelSerializer):
    reporter = ExternalUserSerializer(read_only=True)
    lead = ExternalUserSerializer()
    comms_channel = CommsChannelSerializer(read_only=True)
    action_items = ActionSerializer(read_only=True, many=True)

    class Meta:
        model = Incident
        fields = (
            "action_items",
            "comms_channel",
            "end_time",
            "impact",
            "is_closed",
            "lead",
            "id",
            "report",
            "report_time",
            "reporter",
            "severity",
            "start_time",
            "summary",
        )

    def update(self, instance, validated_data):
        instance.end_time = validated_data.get("end_time", instance.end_time)
        instance.impact = validated_data.get("impact", instance.impact)

        new_lead = validated_data.get("lead", None)
        if new_lead:
            instance.lead = ExternalUser.objects.get(
                display_name=new_lead["display_name"],
                external_id=new_lead["external_id"],
                full_name=new_lead["full_name"],
            )

        instance.report = validated_data.get("report", instance.report)
        instance.start_time = validated_data.get("start_time", instance.start_time)
        instance.summary = validated_data.get("summary", instance.summary)
        instance.severity = validated_data.get("severity", instance.severity)

        instance.save()
        return instance



import pytest

from response.core.models import ExternalUser
from response.slack.cache import get_user_profile, update_user_cache
from tests.slack.slack_payloads import (
    users_list_new,
    users_list_page_1,
    users_list_page_2,
    users_list_response,
)


@pytest.mark.django_db
def test_update_cache_from_empty(mock_slack):
    mock_slack.get_paginated_users.return_value = users_list_response

    update_user_cache()

    assert len(ExternalUser.objects.all()) == 2


@pytest.mark.django_db
def test_update_cache_from_populated(mock_slack):
    mock_slack.get_paginated_users.return_value = users_list_response

    update_user_cache()
    assert len(ExternalUser.objects.all()) == 2

    update_user_cache()
    assert len(ExternalUser.objects.all()) == 2


@pytest.mark.django_db
def test_update_cache_new_user(mock_slack):
    mock_slack.get_paginated_users.side_effect = [users_list_response, users_list_new]

    update_user_cache()
    assert len(ExternalUser.objects.all()) == 2

    update_user_cache()
    assert len(ExternalUser.objects.all()) == 3


@pytest.mark.django_db
def test_update_cache_pagination(mock_slack):
    def get_page(limit=0, cursor=None):
        if not cursor:
            return users_list_page_1
        elif cursor == "page2":
            return users_list_page_2

    mock_slack.get_paginated_users.side_effect = get_page
    update_user_cache()

    users = [user.display_name for user in ExternalUser.objects.all()]

    assert users == ["spengler", "Glinda the Fairly Good"]


@pytest.mark.django_db
def test_get_user_profile_not_in_cache(mock_slack):
    mock_slack.get_user_profile.return_value = {
        "id": "U12345678",
        "name": "spengler",
        "fullname": "Egon Spengler",
    }

    # check cache is empty at start
    assert len(ExternalUser.objects.all()) == 0

    # request a user from cache
    user = get_user_profile("U12345678")

    # check we get back the user from slack
    mock_slack.get_user_profile.assert_called()
    assert user["id"] == "U12345678"

    # check user is now in cache
    assert len(ExternalUser.objects.all()) == 1

    # and that it has the right details populated
    cache_user = ExternalUser.objects.get(external_id="U12345678")
    assert cache_user.display_name == "spengler"
    assert cache_user.full_name == "Egon Spengler"


@pytest.mark.django_db
def test_get_user_profile_in_cache(mock_slack):

    # create cache entry for user
    slack_user = ExternalUser(
        external_id="U12345678", display_name="spengler", full_name="Egon Spengler"
    )
    slack_user.save()
    assert len(ExternalUser.objects.all()) == 1

    # request a user from cache
    user = get_user_profile("U12345678")

    # check we get back the user from cache (i.e. not call to Slack API)
    mock_slack.get_user_profile.assert_not_called()
    assert user["id"] == "U12345678"
    assert user["name"] == "spengler"
    assert user["fullname"] == "Egon Spengler"

    # check cache is unchanged
    assert len(ExternalUser.objects.all()) == 1



import logging
from urllib.parse import urljoin

from django.conf import settings
from django.db import models
from django.urls import reverse

from response.core.models.incident import Incident
from response.slack.client import SlackError

logger = logging.getLogger(__name__)


class CommsChannelManager(models.Manager):
    def create_comms_channel(self, incident):
        """
        Creates a comms channel in slack, and saves a reference to it in the DB
        """
        try:
            name = f"inc-{100+incident.pk}"
            channel_id = settings.SLACK_CLIENT.get_or_create_channel(
                name, auto_unarchive=True
            )
        except SlackError as e:
            logger.error(f"Failed to create comms channel {e}")

        try:
            doc_url = urljoin(
                settings.SITE_URL,
                reverse("incident_doc", kwargs={"incident_id": incident.pk}),
            )

            settings.SLACK_CLIENT.set_channel_topic(
                channel_id, f"{incident.report} - {doc_url}"
            )
        except SlackError as e:
            logger.error(f"Failed to set channel topic {e}")

        comms_channel = self.create(incident=incident, channel_id=channel_id)
        return comms_channel


class CommsChannel(models.Model):

    objects = CommsChannelManager()
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    channel_id = models.CharField(max_length=20, null=False)

    def post_in_channel(self, message: str):
        settings.SLACK_CLIENT.send_message(self.channel_id, message)

    def rename(self, new_name):
        settings.SLACK_CLIENT.rename_channel(self.channel_id, new_name)

    def channel_name(self):
        return settings.SLACK_CLIENT.get_channel_name(self.channel_id)

    def __str__(self):
        return self.incident.report



# Generated by Django 2.2.3 on 2019-09-05 13:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0005_auto_20190906_1413")]

    operations = [
        migrations.AddField(
            model_name="externaluser",
            name="full_name",
            field=models.CharField(blank=True, max_length=50, null=True),
        )
    ]



from django.conf.urls import include, url
from rest_framework import routers

from response.core.views import (
    ActionViewSet,
    ExternalUserViewSet,
    IncidentActionViewSet,
    IncidentsByMonthViewSet,
    IncidentTimelineEventViewSet,
    IncidentViewSet,
)

# Routers provide an easy way of automatically determining the URL conf.
router = routers.DefaultRouter()
router.register(r"incidents", IncidentViewSet, basename="incident")
router.register(
    r"incidents/(?P<incident_pk>[0-9]+)/actions",
    IncidentActionViewSet,
    basename="incident-action",
)
router.register(
    r"incidents/bymonth/(?P<year>[0-9]{4})/(?P<month>[0-9]{2})",
    IncidentsByMonthViewSet,
    basename="incidents-bymonth",
)
router.register(
    r"incidents/(?P<incident_pk>[0-9]+)/timeline/events",
    IncidentTimelineEventViewSet,
    basename="incident-timeline-event",
)
router.register(r"actions", ActionViewSet)
router.register(r"users", ExternalUserViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [url(r"^", include(router.urls))]



users_list_response = {
    "ok": "True",
    "members": [
        {
            "id": "W012A3CDE",
            "team_id": "T012AB3C4",
            "name": "spengler",
            "deleted": "False",
            "color": "9f69e7",
            "real_name": "spengler",
            "tz": "America/Los_Angeles",
            "tz_label": "Pacific Daylight Time",
            "tz_offset": -25200,
            "profile": {
                "avatar_hash": "ge3b51ca72de",
                "status_text": "Print is dead",
                "status_emoji": ":books:",
                "real_name": "Egon Spengler",
                "display_name": "spengler",
                "real_name_normalized": "Egon Spengler",
                "display_name_normalized": "spengler",
                "email": "spengler@ghostbusters.example.com",
                "image_24": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_32": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_48": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_72": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_192": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_512": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "team": "T012AB3C4",
            },
            "is_admin": "True",
            "is_owner": "False",
            "is_primary_owner": "False",
            "is_restricted": "False",
            "is_ultra_restricted": "False",
            "is_bot": "False",
            "updated": 1502138686,
            "is_app_user": "False",
            "has_2fa": "False",
        },
        {
            "id": "U12345678",
            "team_id": "T0G9PQBBK",
            "name": "glinda",
            "deleted": "False",
            "color": "9f69e7",
            "real_name": "Glinda Southgood",
            "tz": "America/Los_Angeles",
            "tz_label": "Pacific Daylight Time",
            "tz_offset": -25200,
            "profile": {
                "avatar_hash": "8fbdd10b41c6",
                "image_24": "https://a.slack-edge.com...png",
                "image_32": "https://a.slack-edge.com...png",
                "image_48": "https://a.slack-edge.com...png",
                "image_72": "https://a.slack-edge.com...png",
                "image_192": "https://a.slack-edge.com...png",
                "image_512": "https://a.slack-edge.com...png",
                "image_1024": "https://a.slack-edge.com...png",
                "image_original": "https://a.slack-edge.com...png",
                "first_name": "Glinda",
                "last_name": "Southgood",
                "title": "Glinda the Good",
                "phone": "",
                "skype": "",
                "real_name": "Glinda Southgood",
                "real_name_normalized": "Glinda Southgood",
                "display_name": "Glinda the Fairly Good",
                "display_name_normalized": "Glinda the Fairly Good",
                "email": "glenda@south.oz.coven",
            },
            "is_admin": "True",
            "is_owner": "False",
            "is_primary_owner": "False",
            "is_restricted": "False",
            "is_ultra_restricted": "False",
            "is_bot": "False",
            "updated": 1480527098,
            "has_2fa": "False",
        },
    ],
    "cache_ts": 1498777272,
    "response_metadata": {"next_cursor": ""},
}


users_list_new = {
    "ok": "True",
    "members": [
        {
            "id": "U00000000",
            "team_id": "T012AB3C4",
            "name": "monzomcmonzoface",
            "deleted": "False",
            "color": "9f69e7",
            "real_name": "spengler",
            "tz": "America/Los_Angeles",
            "tz_label": "Pacific Daylight Time",
            "tz_offset": -25200,
            "profile": {
                "avatar_hash": "ge3b51ca72de",
                "status_text": "Print is dead",
                "status_emoji": ":books:",
                "real_name": "Egon Spengler",
                "display_name": "spengler",
                "real_name_normalized": "Egon Spengler",
                "display_name_normalized": "spengler",
                "email": "spengler@ghostbusters.example.com",
                "image_24": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_32": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_48": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_72": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_192": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_512": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "team": "T012AB3C4",
            },
            "is_admin": "True",
            "is_owner": "False",
            "is_primary_owner": "False",
            "is_restricted": "False",
            "is_ultra_restricted": "False",
            "is_bot": "False",
            "updated": 1502138686,
            "is_app_user": "False",
            "has_2fa": "False",
        }
    ],
    "cache_ts": 1498777272,
    "response_metadata": {"next_cursor": ""},
}

users_list_page_1 = {
    "ok": True,
    "members": [
        {
            "id": "W012A3CDE",
            "team_id": "T012AB3C4",
            "name": "spengler",
            "deleted": False,
            "color": "9f69e7",
            "real_name": "spengler",
            "tz": "America/Los_Angeles",
            "tz_label": "Pacific Daylight Time",
            "tz_offset": -25200,
            "profile": {
                "avatar_hash": "ge3b51ca72de",
                "status_text": "Print is dead",
                "status_emoji": ":books:",
                "real_name": "Egon Spengler",
                "display_name": "spengler",
                "real_name_normalized": "Egon Spengler",
                "display_name_normalized": "spengler",
                "email": "spengler@ghostbusters.example.com",
                "image_24": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_32": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_48": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_72": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_192": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "image_512": "https://.../avatar/e3b51ca72dee4ef87916ae2b9240df50.jpg",
                "team": "T012AB3C4",
            },
            "is_admin": True,
            "is_owner": False,
            "is_primary_owner": False,
            "is_restricted": False,
            "is_ultra_restricted": False,
            "is_bot": False,
            "updated": 1502138686,
            "is_app_user": False,
            "has_2fa": False,
        }
    ],
    "cache_ts": 1498777272,
    "response_metadata": {"next_cursor": "page2"},
}

users_list_page_2 = {
    "ok": True,
    "members": [
        {
            "id": "W07QCRPA4",
            "team_id": "T0G9PQBBK",
            "name": "glinda",
            "deleted": False,
            "color": "9f69e7",
            "real_name": "Glinda Southgood",
            "tz": "America/Los_Angeles",
            "tz_label": "Pacific Daylight Time",
            "tz_offset": -25200,
            "profile": {
                "avatar_hash": "8fbdd10b41c6",
                "image_24": "https://a.slack-edge.com...png",
                "image_32": "https://a.slack-edge.com...png",
                "image_48": "https://a.slack-edge.com...png",
                "image_72": "https://a.slack-edge.com...png",
                "image_192": "https://a.slack-edge.com...png",
                "image_512": "https://a.slack-edge.com...png",
                "image_1024": "https://a.slack-edge.com...png",
                "image_original": "https://a.slack-edge.com...png",
                "first_name": "Glinda",
                "last_name": "Southgood",
                "title": "Glinda the Good",
                "phone": "",
                "skype": "",
                "real_name": "Glinda Southgood",
                "real_name_normalized": "Glinda Southgood",
                "display_name": "Glinda the Fairly Good",
                "display_name_normalized": "Glinda the Fairly Good",
                "email": "glenda@south.oz.coven",
            },
            "is_admin": True,
            "is_owner": False,
            "is_primary_owner": False,
            "is_restricted": False,
            "is_ultra_restricted": False,
            "is_bot": False,
            "updated": 1480527098,
            "has_2fa": False,
        }
    ],
    "cache_ts": 1498777272,
    "response_metadata": {"next_cursor": ""},
}



import factory
from django.contrib.auth.models import User
from faker import Factory

from response.core.models import ExternalUser

faker = Factory.create()


class UserFactory(factory.DjangoModelFactory):
    class Meta:
        model = User

    username = factory.Sequence(lambda i: faker.user_name() + str(i))
    password = factory.LazyFunction(faker.password)


class ExternalUserFactory(factory.DjangoModelFactory):
    class Meta:
        model = ExternalUser

    app_id = "slack"
    display_name = factory.LazyFunction(faker.name)
    external_id = factory.Sequence(lambda i: "U" + str(i).zfill(5))
    owner = factory.SubFactory("tests.factories.UserFactory")



import logging
from urllib.parse import urljoin

from django.conf import settings
from django.db import models
from django.urls import reverse

from response.core.models.incident import Incident
from response.slack.block_kit import Actions, Button, Divider, Message, Section, Text
from response.slack.client import channel_reference, user_reference
from response.slack.decorators.headline_post_action import (
    SLACK_HEADLINE_POST_ACTION_MAPPINGS,
    headline_post_action,
)
from response.slack.models.comms_channel import CommsChannel

logger = logging.getLogger(__name__)


class HeadlinePostManager(models.Manager):
    def create_headline_post(self, incident):
        headline_post = self.create(incident=incident)
        headline_post.update_in_slack()
        return headline_post


class HeadlinePost(models.Model):

    EDIT_INCIDENT_BUTTON = "edit-incident-button"
    CLOSE_INCIDENT_BUTTON = "close-incident-button"
    CREATE_COMMS_CHANNEL_BUTTON = "create-comms-channel-button"
    PAGE_ON_CALL_BUTTON = "page-on-call-button"

    objects = HeadlinePostManager()
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    message_ts = models.CharField(max_length=20, null=True)
    comms_channel = models.OneToOneField(
        CommsChannel, on_delete=models.DO_NOTHING, null=True
    )

    def update_in_slack(self):
        "Creates/updates the slack headline post with the latest incident info"
        logging.info(f"Updating headline post in Slack for incident {self.incident.pk}")
        msg = Message()

        # Set the fallback text so notifications look nice
        msg.set_fallback_text(
            f"{self.incident.report} reported by {user_reference(self.incident.reporter)}"
        )

        # Add report/people
        msg.add_block(
            Section(block_id="report", text=Text(f"*{self.incident.report}*"))
        )
        msg.add_block(
            Section(
                block_id="reporter",
                text=Text(
                    f"🙋🏻‍♂️ Reporter: {user_reference(self.incident.reporter.external_id)}"
                ),
            )
        )
        incident_lead_text = (
            user_reference(self.incident.lead.external_id)
            if self.incident.lead
            else "-"
        )
        msg.add_block(
            Section(
                block_id="lead", text=Text(f"👩‍🚒 Incident Lead: {incident_lead_text}")
            )
        )

        msg.add_block(Divider())

        # Add additional info
        msg.add_block(
            Section(
                block_id="status",
                text=Text(
                    f"{self.incident.status_emoji()} Status: {self.incident.status_text().capitalize()}"
                ),
            )
        )
        severity_text = (
            self.incident.severity_text().capitalize()
            if self.incident.severity_text()
            else "-"
        )
        msg.add_block(
            Section(
                block_id="severity",
                text=Text(
                    f"{self.incident.severity_emoji()} Severity: {severity_text}"
                ),
            )
        )

        doc_url = urljoin(
            settings.SITE_URL,
            reverse("incident_doc", kwargs={"incident_id": self.incident.pk}),
        )
        msg.add_block(
            Section(
                block_id="incident_doc",
                text=Text(f"📄 Document: <{doc_url}|Incident {self.incident.pk}>"),
            )
        )

        channel_ref = (
            channel_reference(self.comms_channel.channel_id)
            if self.comms_channel
            else None
        )
        if channel_ref:
            msg.add_block(
                Section(
                    block_id="comms_channel",
                    text=Text(f"🗣 Comms Channel: {channel_ref or '-'}"),
                )
            )

        # Add buttons (if the incident is open)
        if not self.incident.is_closed():
            msg.add_block(Section(text=Text("Need something else?")))
            actions = Actions(block_id="actions")

            # Add all actions mapped by @headline_post_action decorators
            for key in sorted(SLACK_HEADLINE_POST_ACTION_MAPPINGS.keys()):
                funclist = SLACK_HEADLINE_POST_ACTION_MAPPINGS[key]
                for f in funclist:
                    action = f(self)
                    if action:
                        actions.add_element(action)

            msg.add_block(actions)

        # Post / update the slack message
        response = msg.send(settings.INCIDENT_CHANNEL_ID, self.message_ts)
        logging.info(
            f"Got response back from Slack after updating headline post: {response}"
        )

        # Save the message ts identifier if not already set
        if not self.message_ts:
            self.message_ts = response["ts"]
            self.save()

    def post_to_thread(self, message):
        settings.SLACK_CLIENT.send_message(
            settings.INCIDENT_CHANNEL_ID, message, thread_ts=self.message_ts
        )


# Default/core actions to display on headline post. In order to allow inserting actions between these ones we increment the order by 100


@headline_post_action(order=100)
def create_comms_channel_action(headline_post):
    if headline_post.comms_channel:
        # No need to create an action, channel already exists
        return None
    return Button(
        ":speaking_head_in_silhouette: Create Comms Channel",
        HeadlinePost.CREATE_COMMS_CHANNEL_BUTTON,
        value=headline_post.incident.pk,
    )


@headline_post_action(order=200)
def edit_incident_button(headline_post):
    return Button(
        ":pencil2: Edit",
        HeadlinePost.EDIT_INCIDENT_BUTTON,
        value=headline_post.incident.pk,
    )


@headline_post_action(order=300)
def close_incident_button(headline_post):
    return Button(
        ":white_check_mark: Close",
        HeadlinePost.CLOSE_INCIDENT_BUTTON,
        value=headline_post.incident.pk,
    )



import logging

import after_response

from response.core.models.incident import Incident
from response.slack.models.comms_channel import CommsChannel

logger = logging.getLogger(__name__)

SLACK_ACTION_MAPPINGS = {}


class ActionContext:
    def __init__(
        self, incident, user_id, message, button_value, trigger_id, response_url
    ):
        self.incident = incident
        self.user_id = user_id
        self.message = message
        self.button_value = button_value
        self.trigger_id = trigger_id
        self.response_url = response_url


def action_handler(callback_id, func=None):
    def _wrapper(fn):
        SLACK_ACTION_MAPPINGS[callback_id] = fn
        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def remove_action_handler(callback_id):
    SLACK_ACTION_MAPPINGS.pop(callback_id, None)


@after_response.enable
def handle_action(payload):
    actions = payload["actions"]
    if actions:
        action_id = actions[0]["action_id"]

        if action_id not in SLACK_ACTION_MAPPINGS:
            logger.error(f"Can't find handler for action id {action_id}")
            return

        handler = SLACK_ACTION_MAPPINGS[action_id]

        user_id = payload["user"]["id"]
        channel_id = payload["channel"]["id"]
        button_value = actions[0]["value"]
        message = payload["message"]
        trigger_id = payload["trigger_id"]
        response_url = payload["response_url"]

        # we want to tie all actions to an incident, and have two ways to do this:
        # - if action comes from a comms channel, lookup the incident by comms channel id
        # - if not in comms channel, we rely on the button value containing the incident id
        try:
            comms_channel = CommsChannel.objects.get(channel_id=channel_id)
            incident = comms_channel.incident
        except CommsChannel.DoesNotExist:
            incident_id = button_value
            incident = Incident.objects.get(pk=incident_id)
        except Incident.DoesNotExist:
            logger.error(
                f"Can't find incident associated with channel {channel_id} or with id {incident_id}"
            )
            return

        action_context = ActionContext(
            incident=incident,
            user_id=user_id,
            message=message,
            button_value=button_value,
            trigger_id=trigger_id,
            response_url=response_url,
        )

        handler(action_context)



import json
import logging

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from response.core.models.incident import Incident
from response.slack.authentication import slack_authenticate
from response.slack.cache import update_user_cache
from response.slack.decorators import (
    handle_action,
    handle_dialog,
    handle_event,
    handle_notifications,
)
from response.slack.dialog_builder import (
    Dialog,
    SelectFromUsers,
    SelectWithOptions,
    Text,
    TextArea,
)
from response.slack.settings import INCIDENT_REPORT_DIALOG

logger = logging.getLogger(__name__)


@csrf_exempt
@slack_authenticate
def slash_command(request):
    """
    Handles slash commands from slack
    More details here: https://api.slack.com/slash-commands

    @param request the request from slack containing the slash command
    @return: return a HTTP response to indicate the request was handled
    """

    user_id = request.POST.get("user_id")
    trigger_id = request.POST.get("trigger_id")
    report = request.POST.get("text")

    dialog = Dialog(
        title="Report an Incident",
        submit_label="Report",
        elements=[
            Text(
                label="Report",
                name="report",
                placeholder="What's the tl;dr?",
                value=report,
            ),
            TextArea(
                label="Summary",
                name="summary",
                optional=True,
                placeholder="Can you share any more details?",
            ),
            TextArea(
                label="Impact",
                name="impact",
                optional=True,
                placeholder="Who or what might be affected?",
                hint="Think about affected people, systems, and processes",
            ),
            SelectFromUsers(label="Lead", name="lead", optional=True),
            SelectWithOptions(
                [(s.capitalize(), i) for i, s in Incident.SEVERITIES],
                label="Severity",
                name="severity",
                optional=True,
            ),
        ],
    )

    logger.info(
        f"Handling Slack slash command for user {user_id}, report {report} - opening dialog"
    )

    dialog.send_open_dialog(INCIDENT_REPORT_DIALOG, trigger_id)
    return HttpResponse()


@csrf_exempt
@slack_authenticate
def action(request):
    """
    Handles actions sent from Slack.

    @param request the request from slack containing an action
    @return: return a HTTP response to indicate the request was handled
    """
    payload = json.loads(request.POST["payload"])
    action_type = payload["type"]

    logger.info(f"Handling Slack action of type '{action_type}'")

    if action_type == "dialog_submission":
        handle_dialog.after_response(payload)
    elif action_type == "block_actions":
        handle_action.after_response(payload)
    else:
        logger.error(f"No handler for action type {action_type}")

    return HttpResponse()


@csrf_exempt
@slack_authenticate
def event(request):
    """
    Handles events sent from Slack.

    Details: We can configure our Slack app receive specific events, and doing so tells
    slack to call this endpoint with a JSON payload representing the event.

    See here for reference: https://api.slack.com/events-api

    @param request the request from slack containing an event
    @return: return a HTTP response to indicate the request was handled
    """
    payload = json.loads(request.body)
    action_type = payload["type"]

    logger.info(f"Handling Slack event of type '{action_type}'")

    if action_type == "event_callback":
        handle_event(payload)
    elif action_type == "url_verification":
        # the url_verification event is called when we change the registered event callback url
        # in the Sl ack app configuration.  It expects us to return the challenge token sent in
        # the request
        return HttpResponse(payload["challenge"])

    return HttpResponse()


@csrf_exempt
def cron_minute(request):
    "Handles actions that need to take place every minute"
    handle_notifications()
    return HttpResponse()


@csrf_exempt
def cron_daily(request):
    "Handles actions that need to take place every day"
    update_user_cache()
    return HttpResponse()



# Generated by Django 2.2.3 on 2019-07-19 10:27

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="CommsChannel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("channel_id", models.CharField(max_length=20)),
            ],
        ),
        migrations.CreateModel(
            name="ExternalUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app_id", models.CharField(max_length=50)),
                ("external_id", models.CharField(max_length=50)),
                ("display_name", models.CharField(max_length=50)),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"unique_together": {("owner", "app_id")}},
        ),
        migrations.CreateModel(
            name="Incident",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("report", models.CharField(max_length=200)),
                ("report_time", models.DateTimeField()),
                ("start_time", models.DateTimeField()),
                ("end_time", models.DateTimeField(blank=True, null=True)),
                (
                    "summary",
                    models.TextField(
                        blank=True,
                        help_text="What's the high level summary?",
                        null=True,
                    ),
                ),
                (
                    "impact",
                    models.TextField(
                        blank=True, help_text="What impact is this having?", null=True
                    ),
                ),
                (
                    "severity",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("1", "critical"),
                            ("2", "major"),
                            ("3", "minor"),
                            ("4", "trivial"),
                        ],
                        max_length=10,
                        null=True,
                    ),
                ),
                (
                    "lead",
                    models.ForeignKey(
                        blank=True,
                        help_text="Who is leading?",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="lead",
                        to="response.ExternalUser",
                    ),
                ),
                (
                    "reporter",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="reporter",
                        to="response.ExternalUser",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PinnedMessage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("message_ts", models.CharField(max_length=50)),
                ("text", models.TextField()),
                ("timestamp", models.DateTimeField()),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="response.ExternalUser",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="HeadlinePost",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("message_ts", models.CharField(max_length=20, null=True)),
                (
                    "comms_channel",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to="response.CommsChannel",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="commschannel",
            name="incident",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="response.Incident"
            ),
        ),
        migrations.CreateModel(
            name="Action",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("details", models.TextField(blank=True, default="")),
                ("done", models.BooleanField(default=False)),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.ExternalUser",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserStats",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("join_time", models.DateTimeField(null=True)),
                ("message_count", models.IntegerField(default=0)),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.ExternalUser",
                    ),
                ),
            ],
            options={"unique_together": {("incident", "user")}},
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=30)),
                ("time", models.DateTimeField()),
                ("repeat_count", models.IntegerField(default=0)),
                ("completed", models.BooleanField(default=False)),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
            ],
            options={"unique_together": {("incident", "key")}},
        ),
        migrations.CreateModel(
            name="IncidentExtension",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=50)),
                ("value", models.CharField(max_length=100)),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
            ],
            options={"unique_together": {("incident", "key")}},
        ),
    ]



# Generated by Django 2.2.3 on 2019-08-19 15:07

import datetime

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("response", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="TimelineEvent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "timestamp",
                    models.DateTimeField(
                        default=datetime.datetime.now,
                        help_text="Time of when this event occurred.",
                    ),
                ),
                (
                    "text",
                    models.TextField(help_text="Freeform text describing this event"),
                ),
                (
                    "event_type",
                    models.CharField(
                        choices=[("text", "Freeform text field")],
                        help_text="Type of event.",
                        max_length=10,
                    ),
                ),
                (
                    "metadata",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text="Additional fields that can be added by other event types",
                        null=True,
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="response.Incident",
                    ),
                ),
            ],
        )
    ]



from .action import Action
from .incident import Incident
from .timeline import TimelineEvent, add_incident_update_event
from .user_external import ExternalUser

__all__ = (
    "Action",
    "Incident",
    "TimelineEvent",
    "ExternalUser",
    "add_incident_update_event",
)



from django.contrib.auth.models import User
from django.db import models


class ExternalUserManager(models.Manager):
    def get_or_create_slack(self, *args, **kwargs):
        return self.get_or_create(app_id="slack", *args, **kwargs)

    def update_or_create_slack(self, *args, **kwargs):
        return self.update_or_create(app_id="slack", *args, **kwargs)


class ExternalUser(models.Model):
    class Meta:
        unique_together = ("owner", "app_id")

    owner = models.ForeignKey(User, on_delete=models.PROTECT, null=True, blank=True)
    app_id = models.CharField(max_length=50, blank=False, null=False)
    external_id = models.CharField(max_length=50, blank=False, null=False)
    display_name = models.CharField(max_length=50, blank=False, null=False)
    full_name = models.CharField(max_length=50, blank=True, null=True)

    objects = ExternalUserManager()

    def __str__(self):
        return f"{self.display_name or self.external_id} ({self.app_id})"



import logging
from collections import defaultdict

from response.slack.models.comms_channel import CommsChannel

logger = logging.getLogger(__name__)

# Stores a map from slack event name to a list of callback functions
EVENT_MAPPINGS = defaultdict(list)


def slack_event(event, func=None):
    """
    @slack_event is a decorator which registers a function as a handler
    for a particular slack_event (e.g. app_mention, pin_added, etc.)

    Arguments:
        event: Command to invoke this on

    Example usage:

    @slack_event('pin_added')
    def handle_pin_added(event_payload):
        do_some_stuff()
    """

    def _wrapper(fn):
        EVENT_MAPPINGS[event].append(fn)
        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def handle_event(payload):
    """
    Handles slack event callbacks and routes the action to the correct event handler

    @param payload the Slack event JSON payload
    """

    # the actual event is nested inside the 'event' key of the payload
    event = payload.get("event", "")
    event_type = event.get("type", "")

    logger.info(f"Handling Slack event {event} of type {event_type}")

    # ignore bot messages
    if event.get("subtype", None) == "bot_message":
        logger.info(f"Ignoring bot message")
        return

    # if it doesn't exist, error and return
    if event_type not in EVENT_MAPPINGS:
        logger.error(f"No handler found for event <{event_type}>")
        return

    # events use either channel_id _or_ channel as the key (thanks Slack)
    if "channel_id" in event:
        channel_id = event["channel_id"]
    elif "channel" in event:
        channel_id = event["channel"]
    else:
        channel_id = None

    # get the incident by the comms_channel_id
    try:
        comms_channel = CommsChannel.objects.get(channel_id=channel_id)
        incident = comms_channel.incident
    except CommsChannel.DoesNotExist:
        logger.error(f"Can't find incident associated with channel_id {channel_id}")
        return

    # call the registered handler
    for handler in EVENT_MAPPINGS[event_type]:
        logger.info(
            f"Calling handler for event type {event_type} for incident {incident.pk}"
        )
        handler(incident, payload["event"])



import json

from django.urls import reverse
from rest_framework.test import force_authenticate

from response.core.views import ExternalUserViewSet
from tests.factories import ExternalUserFactory


def test_list_users(arf, api_user):
    users = ExternalUserFactory.create_batch(5)

    req = arf.get(reverse("externaluser-list"))
    force_authenticate(req, user=api_user)
    response = ExternalUserViewSet.as_view({"get": "list"})(req)

    assert response.status_code == 200, "Got non-200 response from API"
    content = json.loads(response.rendered_content)

    assert "results" in content, "Response didn't have results key"

    assert len(content["results"]) == (
        len(users) + 1
    )  # account for the API user we created in a fixture

    for user in content["results"]:
        assert user["app_id"]
        assert user["external_id"]
        assert user["display_name"]



import logging

from django.conf import settings

from response.slack.client import SlackError
from response.slack.models.comms_channel import CommsChannel

logger = logging.getLogger(__name__)

COMMAND_MAPPINGS = {}
COMMAND_MAPPINGS_CUSTOM = {}
COMMAND_HELP = {}


def get_help():
    """
    get_help returns the help string for a command
    """
    rendered = ""
    for k in sorted(COMMAND_HELP.keys()):
        rendered += f"`{k}` -  {COMMAND_HELP[k]}\n"
    return rendered


def get_commands():
    return COMMAND_MAPPINGS.keys()


def __default_incident_command(commands: list, func=None, helptext=""):
    """
    @__default_incident_command is a decorator which registers a function as
    the default handler for a set of command strings

    Arguments:
        commands: A list of strings to register as commands
        helptext: Text to be displayed when the "help" command is run

    Example usage:

    @__default_incident_command('lead', helptext='Set the incident lead')
    def handle_incident_lead(context):
        do_some_stuff()
    """

    def _wrapper(fn):
        for command in commands:
            COMMAND_MAPPINGS[command] = fn

        COMMAND_HELP[", ".join(commands)] = helptext

        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def incident_command(commands: list, func=None, helptext=""):
    """
    @incident_command is a decorator which registers a function as a handler
    for a set of command strings

    Arguments:
        commands: A list of strings to register as commands
        helptext: Text to be displayed when the "help" command is run

    Example usage:

    @incident_command('lead', helptext='Set the incident lead')
    def handle_incident_lead(context):
        do_some_stuff()
    """

    def _wrapper(fn):
        for command in commands:
            COMMAND_MAPPINGS_CUSTOM[command] = fn

        COMMAND_HELP[", ".join(commands)] = helptext

        return fn

    if func:
        return _wrapper(func)
    return _wrapper


def handle_incident_command(command_name, message, thread_ts, channel_id, user_id):
    """
    Handler for app mention events of the format @<bot> <command> <extra text>

    @param payload an app mention string of the form @incident summary Something's happened
    """
    if (
        command_name not in COMMAND_MAPPINGS
        and command_name not in COMMAND_MAPPINGS_CUSTOM
    ):
        settings.SLACK_CLIENT.send_ephemeral_message(
            channel_id,
            user_id,
            "I'm sorry, I don't know how to help with that :grimacing:",
        )

        react_not_ok(channel_id, thread_ts)

        logger.error(f"No handler found for command {command_name}")
        return

    if command_name in COMMAND_MAPPINGS_CUSTOM:
        logger.info(
            f"Handling incident command {command_name} '{message}' in channel {channel_id} with custom handler"
        )
        command = COMMAND_MAPPINGS_CUSTOM[command_name]
    else:
        logger.info(
            f"Handling incident command {command_name} '{message}' in channel {channel_id}"
        )
        command = COMMAND_MAPPINGS[command_name]

    try:
        comms_channel = CommsChannel.objects.get(channel_id=channel_id)
        handled, response = command(comms_channel.incident, user_id, message)

        if handled:
            react_ok(channel_id, thread_ts)
        else:
            react_not_ok(channel_id, thread_ts)

        if response:
            settings.SLACK_CLIENT.send_message(comms_channel.channel_id, response)

    except CommsChannel.DoesNotExist:
        logger.error("No matching incident found for this channel")
    except Exception as e:
        logger.error(f"Error handling incident command {command_name} {message}: {e}")
        raise


def react_not_ok(channel_id, thread_ts):
    try:
        settings.SLACK_CLIENT.remove_reaction("white_check_mark", channel_id, thread_ts)
    except SlackError as e:
        logger.error(
            f"Couldn't remove existing reaction from {channel_id} - {thread_ts}. Error: {e}"
        )

    try:
        settings.SLACK_CLIENT.add_reaction("question", channel_id, thread_ts)
    except SlackError as e:
        logger.error(
            f"Couldn't add 'question' reaction to {channel_id} - {thread_ts}. Error: {e}"
        )


def react_ok(channel_id, thread_ts):
    try:
        settings.SLACK_CLIENT.remove_reaction("question", channel_id, thread_ts)
    except SlackError as e:
        logger.error(
            f"Couldn't remove existing reaction from {channel_id} - {thread_ts}. Error: {e}"
        )

    try:
        settings.SLACK_CLIENT.add_reaction("white_check_mark", channel_id, thread_ts)
    except SlackError as e:
        logger.error(
            f"Couldn't add 'white_check_mark' reaction to {channel_id} - {thread_ts}. Error: {e}"
        )



import os
from os import path

from setuptools import find_packages, setup

VERSION = "0.1.7"

INSTALL_REQUIRES = [
    "Django>=2.2",
    "bleach==3.1.0",
    "cryptography>=2.7",
    "django-after-response>=0.2.2",
    "django-bootstrap4>=0.0.7",
    "djangorestframework>=3.9.2",
    "emoji-data-python==1.1.0",
    "jsonfield>=2.0.2",
    "markdown2>=2.3.7",
    "python-slugify>=1.2.6",
    "slackclient>=1.3,<2",
    "statuspageio>=0.0.1",
]

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# load README.md
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="django-incident-response",
    version=VERSION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude="demo"),
    install_requires=INSTALL_REQUIRES,
    package_dir={"response": "response"},
    python_requires=">3.6",
    include_package_data=True,
    license="MIT License",  # example license
    description="A real-time incident response and reporting tool",
    url="https://github.com/monzo/response",
    author="Chris Evans",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import logging

import emoji_data_python
from django import template
from django.conf import settings

from response.slack.client import slack_to_human_readable

register = template.Library()

logger = logging.getLogger(__name__)


@register.filter
def unslackify(value):
    """Takes a string and turns slack style :emoji: into unicode"""

    value = slack_to_human_readable(value)

    # replace all colon style :emoji: with unicode
    value = emoji_data_python.replace_colons(value)

    return value


@register.filter
def slack_id_to_fullname(value):
    profile = settings.SLACK_CLIENT.get_user_profile(value)
    if profile:
        return profile["fullname"]



from urllib.parse import urljoin

from django.conf import settings
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from django.urls import reverse

from response.core.models import Incident, add_incident_update_event
from response.core.serializers import ExternalUserSerializer
from response.slack.models import HeadlinePost


@receiver(post_save, sender=Incident)
def update_headline_after_incident_save(sender, instance, **kwargs):
    """
    Reflect changes to incidents in the headline post

    Important: this is called in the synchronous /incident flow so must remain fast (<2 secs)

    """
    try:
        headline_post = HeadlinePost.objects.get(incident=instance)
        headline_post.update_in_slack()

    except HeadlinePost.DoesNotExist:
        headline_post = HeadlinePost.objects.create_headline_post(incident=instance)


@receiver(pre_save, sender=Incident)
def prompt_incident_report(sender, instance: Incident, **kwargs):
    """
    Prompt incident lead to complete a report when an incident is closed.
    """

    try:
        prev_state = Incident.objects.get(pk=instance.pk)
    except Incident.DoesNotExist:
        # Incident hasn't been saved yet, nothing to do here.
        return

    if instance.is_closed() and not prev_state.is_closed():
        user_to_notify = instance.lead or instance.reporter
        doc_url = urljoin(
            settings.SITE_URL,
            reverse("incident_doc", kwargs={"incident_id": instance.pk}),
        )
        settings.SLACK_CLIENT.send_message(
            user_to_notify.external_id,
            f"👋 Don't forget to fill out an incident report here: {doc_url}",
        )


@receiver(post_save, sender=HeadlinePost)
def update_headline_after_save(sender, instance, **kwargs):
    """
    Reflect changes to headline posts in slack

    """
    instance.update_in_slack()


@receiver(pre_save, sender=Incident)
def add_timeline_events(sender, instance: Incident, **kwargs):
    try:
        prev_state = Incident.objects.get(pk=instance.pk)
    except Incident.DoesNotExist:
        # Incident hasn't been saved yet, nothing to do here.
        return

    if prev_state.lead != instance.lead:
        update_incident_lead_event(prev_state, instance)

    if prev_state.report != instance.report:
        update_incident_report_event(prev_state, instance)

    if prev_state.summary != instance.summary:
        update_incident_summary_event(prev_state, instance)

    if prev_state.impact != instance.impact:
        update_incident_impact_event(prev_state, instance)

    if prev_state.severity != instance.severity:
        update_incident_severity_event(prev_state, instance)


def update_incident_lead_event(prev_state, instance):
    old_lead = None
    if prev_state.lead:
        old_lead = ExternalUserSerializer(prev_state.lead).data

    new_lead = None
    if instance.lead:
        new_lead = ExternalUserSerializer(instance.lead).data

    if prev_state.lead:
        if instance.lead:
            text = f"Incident lead changed from {prev_state.lead.display_name} to {instance.lead.display_name}"
        else:
            text = f"{prev_state.lead.display_name} was removed as incident lead"

    else:
        text = f"{instance.lead.display_name} was added as incident lead"

    add_incident_update_event(
        incident=instance,
        update_type="incident_lead",
        text=text,
        old_value=old_lead,
        new_value=new_lead,
    )


def update_incident_report_event(prev_state, instance):
    add_incident_update_event(
        incident=instance,
        update_type="incident_report",
        text=f'Incident report updated from "{prev_state.report}" to "{instance.report}"',
        old_value=prev_state.report,
        new_value=instance.report,
    )


def update_incident_summary_event(prev_state, instance):
    if prev_state.summary:
        text = (
            f'Incident summary updated from "{prev_state.summary}" to "{instance.summary}"',
        )
    else:
        text = f'Incident summary added: "{instance.summary}"'

    add_incident_update_event(
        incident=instance,
        update_type="incident_summary",
        text=text,
        old_value=prev_state.summary,
        new_value=instance.summary,
    )


def update_incident_impact_event(prev_state, instance):
    if prev_state.impact:
        text = (
            f'Incident impact updated from "{prev_state.impact}" to "{instance.impact}"',
        )
    else:
        text = f'Incident impact added: "{instance.impact}"'

    add_incident_update_event(
        incident=instance,
        update_type="incident_impact",
        text=text,
        old_value=prev_state.impact,
        new_value=instance.impact,
    )


def update_incident_severity_event(prev_state, instance):
    if prev_state.severity:
        text = (
            f"Incident severity updated from {prev_state.severity_text()} to {instance.severity_text()}",
        )
    else:
        text = f"Incident severity set to {instance.severity_text()}"

    add_incident_update_event(
        incident=instance,
        update_type="incident_severity",
        text=text,
        old_value={
            "id": prev_state.severity,
            "text": prev_state.severity_text() if prev_state else "",
        },
        new_value={"id": instance.severity, "text": instance.severity_text()},
    )
